import tensorflow as tf
from tensorflow import keras


class LinearHyperbolic(keras.layers.Layer):
    """
    Implementation of a hyperbolic linear layer for a neural network, that inherits from the keras Layer class
    """

    def __init__(self, units, manifold, c, activation=None, use_bias=True):
        super().__init__()
        self.units = units
        self.c = tf.Variable([c], dtype="float64")
        self.manifold = manifold
        self.activation = keras.activations.get(activation)
        self.use_bias = use_bias

    def build(self, batch_input_shape):
        w_init = tf.random_normal_initializer()
        self.kernel = tf.Variable(
            initial_value=w_init(shape=(batch_input_shape[-1], self.units), dtype="float64"), dtype="float64", trainable=True,
        )

        if self.use_bias:
            self.bias = self.add_weight(
                name="bias",
                shape=(self.units),
                initializer="zeros",
                dtype=tf.float64,
                trainable=True,
            )

        super().build(batch_input_shape)  # must be at the end

    def call(self, inputs):
        """
        Called during forward pass of a neural network. Uses hyperbolic matrix multiplication
        """
        # TODO: remove casting and instead recommend setting default tfd values to float64
        inputs = tf.cast(inputs, tf.float64)
        mv = self.manifold.mobius_matvec(self.kernel, inputs, self.c)
        res = self.manifold.proj(mv, c=self.c)

        if self.use_bias:
            hyp_bias = self.manifold.expmap0(self.bias, c=self.c)
            hyp_bias = self.manifold.proj(hyp_bias, c=self.c)
            res = self.manifold.mobius_add(res, hyp_bias, c=self.c)
            res = self.manifold.proj(res, c=self.c)

        return self.activation(res)

    def get_config(self):
        base_config = super().get_config()
        return {
            **base_config,
            "units": self.units,
            "activation": keras.activations.serialize(self.activation),
            "manifold": self.manifold,
            "curvature": self.c
        }



'''
This package contains functions for high precision calculations of hyperbolic functions.
Note that a hyperbolic distance of d requires O(d) bits of precision to compute.
'''
import mpmath as mpm
import numpy as np

def poincare_dist(x, y, c=1.0, precision=None):
    ''' 
    The hyperbolic distance between points in the Poincare model with curvature -1/c 
        Args:
            x, y: size 1xD mpmath matrix representing point in the D-dimensional ball |x| < 1
            precision (int): bits of precision to use
        Returns:
            mpmath float object. Can be converted back to regular float
    '''
    if precision is not None:
        mpm.mp.dps = precision
    x2 = mpm.fdot(x,x) 
    y2 = mpm.fdot(y,y) 
    xy = mpm.fdot(x,y)
    sqrt_c = mpm.sqrt(c)
    denom = 1 - 2*c*xy + c**2 * x2*y2
    norm = mpm.norm( ( -(1-2*c*xy + c*y2)*x + (1.-c*x2)*y ) /denom )
    return 2/sqrt_c * mpm.atanh( sqrt_c * norm ) 

def poincare_dist0(x, c=1.0, precision=None):
    ''' Distance from 0 to x in the Poincare model with curvature -1/c'''
    if precision is not None:
        mpm.mp.dps = precision
    x_norm = mpm.norm(x)
    sqrt_c = mpm.sqrt(c)
    return 2/sqrt_c * mpm.atanh( x_norm / sqrt_c)

def poincare_metric(X, precision=None):
    ''' Calculate the distance matrix of points in the Poincare model with curvature -1/c 
        Args:
            X (ndarray): N x D matrix representing N points in the D-dimensional ball |x|< 1
            precision (int): bits of precision to use
        Returns:
            distance matrix in compressed 1D form ( see scipy.squareform )
    '''
    if precision is not None:
        mpm.mp.dps = precision
    N = X.shape[0]
    out = np.zeros(shape=(N*(N-1)//2), dtype=np.float64)
    for i in range(N):
        idx = N*i-((i+1)*(i+2))//2
        for j in range(i+1,N):
            out[idx+j] = poincare_dist(X[i],X[j],precision=precision)
    return out

def poincare_reflect(a, x, c=1.0, precision=None):
    '''
    Spherical inversion (or "Poincare reflection") of a point x about a sphere
    with center at point "a", and radius = 1/c.
    '''
    if precision is not None:
        mpm.mp.dps = precision
    a2 = mpm.fdot(a,a)
    x2 = mpm.fdot(x,x)
    xa2 = x2 + a2 - 2*mpm.fdot(x,a)
    r2 = a2 - 1./c 
    scal = mpm.fdiv(r2, xa2)
    return scal*(x-a) + a

def poincare_reflect0(z, x, c=1.0, precision=None):
    '''
    Spherical inversion (or "Poincare reflection") of a point x
    such that point z is mapped to the origin.
    '''
    if precision is not None:
        mpm.mp.dps = precision
    # the reflection is poincare_reflect(a, x) where
    # a = c * z / |z|**2
    z2 = mpm.fdot(z,z)
    zscal = c / z2
    x2 = mpm.fdot(x,x)
    a2 = c* zscal
    r2 = a2 - 1./c
    xa2 = x2 + a2 - 2*zscal*mpm.fdot(z,x)
    scal = mpm.fdiv(r2, xa2)
    return scal*( x - zscal * z) + zscal * z



import numpy as np
from scipy.spatial.distance import squareform
from random import randint

# there are more efficient algorithms for this
# https://people.csail.mit.edu/virgi/6.890/papers/APBP.pdf
def max_min(A, B):
    '''max-min product of two square matrices
    params: 
        A, B: NxN numpy arrays '''
    assert A.shape == B.shape
    return np.max(np.minimum(A[:, :, None], B[None, :, :]), axis=1)

def mat_gromov_prod(dists, base):
    '''Gromov products of N-point metric space relative to base point
    Args:
        dists (ndarray): NxN matrix of pairwise distances
        base (int): index of the basepoint in 0...N-1 '''
    assert dists.shape[0] == dists.shape[1] and 0 <= base < dists.shape[0]
    row = dists[base, :][None, :]
    col = dists[:, base][:, None]
    return 0.5*(row+col-dists)

def delta_rel(dists, base=None):
    ''' Measure the delta-hyperbolicity constant of data 
    with respect to basepoint, normalized by the diameter (max dist).
    Args:
        dists (ndarray): NxN matrix of pairwise distances
        base (int): index of basepoint in 0...N-1 (default = random)
    '''
    if base is None:
        base = randint(0,dists.shape[0]-1)
    assert is_metric(dists) and 0 <= base < dists.shape[0]
    G = mat_gromov_prod(dists, base)
    delta = np.max(max_min(G,G)-G)
    diam = np.max(dists)
    return delta/diam

def delta_sample(X, **kwargs):
    bs = kwargs.get("bs", X.shape[0])
    tries = kwargs.get("tries", 10)
    dist = kwargs.get("dist", None)
    deltas = []
    for i in range(tries):
        idx = np.random.choice(X.shape[0], bs)
        batch = X[idx]
        if dist is None:
            dists = np.linalg.norm(
                        batch[None:,]-batch[:,None],
                        axis=-1)
        else:
            dists = dist(batch,batch)
        deltas.append(
                delta_rel(dists,randint(0,bs-1))
                )
    return deltas 

def is_metric(X, tol=1e-8):
    return len(X.shape) == 2 and \
            np.all( np.abs(X-X.T)<tol ) and\
            np.all( np.abs(np.diag(X))<tol ) and\
            np.all(X >= 0)

def avg_distortion(metric1, metric2):
    ''' Average distortion between two metrics.
        Args:
            metric1, metric2 (ndarray): N x N distance matrices, 
                    or length N*(N-1)//2 compressed distance matrices
        Returns:
            average distortion (float)
    ''' 
    assert metric1.shape == metric2.shape
    if len(metric1.shape) > 1:
        assert is_metric(metric1)
        X = squareform(metric1)
    else:
        X = metric1
    if len(metric2.shape) > 1:
        assert is_metric(metric2)
        Y = squareform(metric2)
    else:
        Y = metric2
    return np.mean( np.abs(X-Y)/Y )



import tensorflow as tf
from tensorflow import keras

class HypLuongAttention(keras.layers.Attention):
  """Dot-product attention layer, a.k.a. Luong-style attention.
  Inputs are `query` tensor of shape `[batch_size, Tq, dim]`, `value` tensor of
  shape `[batch_size, Tv, dim]` and `key` tensor of shape

  Call Args:
    inputs: List of the following tensors:
      * query: Query `Tensor` of shape `[batch_size, Tq, dim]`.
      * value: Value `Tensor` of shape `[batch_size, Tv, dim]`.
  Output:
    Attention outputs of shape `[batch_size, Tq, dim]`.
  # ...
  ```
  """

  def __init__(self, manifold, c=1, use_scale=False, hyperbolic_input=False, **kwargs):
    super(HypLuongAttention, self).__init__(**kwargs)
    self.use_scale = use_scale
    self.hyperbolc_input = hyperbolic_input
    self.manifold = manifold
    self.c = tf.Variable([c], dtype="float64")

  def build(self, input_shape):
    """Creates scale variable if use_scale==True."""
    if self.use_scale:
      self.scale = self.add_weight(
          name='scale',
          shape=(),
          initializer='ones',
          dtype=self.dtype,
          trainable=True)
    else:
      self.scale = None
    super(HypLuongAttention, self).build(input_shape)

  def _calculate_scores(self, query, key):
    """Calculates attention scores as a query-key dot product.
    Args:
      query: Query tensor of shape `[batch_size, Tq, dim]`.
      key: Key tensor of shape `[batch_size, Tv, dim]`.
    Returns:
      Tensor of shape `[batch_size, Tq, Tv]`.
    """
    if self.hyperbolc_input:
        scores = self.manifold.single_query_attn_scores(query, key, self.c)
    else:
        scores = tf.linalg.matmul(query, key, transpose_b=True)

        if self.scale is not None:
          scores *= self.scale

    return scores

  def get_config(self):
    config = {'use_scale': self.use_scale}
    base_config = super(keras.layers.Attention, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))



import networkx as nx
import numpy as np
import scipy.sparse as sp

def to_networkx(edge_weights):
    ''' Convert edge weight dict to networkx weighted graph'''
    G = nx.Graph()
    for e in edge_weights:
        G.add_edge(e[0], e[1], weight=edge_weights[e])
    return G

def to_sparse(edge_weights):
    ''' Convert edge weight dict to compressed adjacency matrix,
    stored in a scipy.sparse.dok_matrix
    '''
    N = max([e[1] for e in edge_weights])+1
    mat = sp.dok_matrix((N,N), dtype=np.float64)
    for e in edge_weights:
        mat[e[0],e[1]] = edge_weights[e]
        mat[e[1],e[0]] = edge_weights[e]
    return mat

def binary_tree(depth):
    return n_ary_tree(2, depth)

def trinary_tree(depth):
    return n_ary_tree(3, depth) 

def n_ary_tree(n, depth):
    assert n >= 2 
    assert depth >= 0 
    T = nx.Graph()
    for d in range(depth):
        for i in range(n**d,n**(d+1)):
            for j in range(n):
                T.add_edge(i, n*i+j)
    return nx.relabel.convert_node_labels_to_integers(T)



import tensorflow as tf
from tensorflow import keras

class RSGD(keras.optimizers.Optimizer):

    """
    Implmentation of a Riemannian Stochastic Gradient Descent. This class inherits form the keras Optimizer class.
    """

    def __init__(self, learning_rate=0.01, name="SGOptimizer", **kwargs):
        """Call super().__init__() and use _set_hyper() to store hyperparameters"""
        super().__init__(name, **kwargs)
        self._set_hyper(
            "learning_rate", kwargs.get("lr", tf.cast(learning_rate, tf.float64))
        )  # handle lr=learning_rate
        self._is_first = True

    def _create_slots(self, var_list):
        """
        For each model variable, create the optimizer variable associated with it
        """
        for var in var_list:
            self.add_slot(var, "pv")  # previous variable i.e. weight or bias
        for var in var_list:
            self.add_slot(var, "pg")  # previous gradient

    @tf.function
    def _resource_apply_dense(self, grad, var):
        """
        Update the slots and perform one optimization step for one model variable
        """
        r_grad = self.rgrad(var, grad)
        r_grad = tf.math.multiply(r_grad, -self.lr)
        new_var_m = self.expm(var, r_grad)

        # slots aren't currently used - they store previous weights and gradients
        pv_var = self.get_slot(var, "pv")
        pg_var = self.get_slot(var, "pg")
        pv_var.assign(var)
        pg_var.assign(grad)
        # assign new weights

        var.assign(new_var_m)

    def rgrad(self, var, grads):
        """
        Transforms the gradients to Riemannian space
        """
        vars_sqnorm = tf.math.reduce_sum(var ** 2, axis=-1, keepdims=True)
        grads = grads * tf.broadcast_to(((1 - vars_sqnorm) ** 2 / 4), tf.shape(grads))
        return grads

    def expm(self, p, d_p, normalize=False, lr=None, out=None):
        """
        Maps the variable values
        """
        if lr is not None:
            d_p = tf.math.multiply(d_p, -lr)
        if out is None:
            out = p
        p = tf.math.add(p, d_p)
        if normalize:
            self.normalize(p)
        return p

    def normalize(self, u):
        d = u.shape[-1]
        if self.max_norm:
            u = tf.reshape(u, [-1, d])
            u.renorm_(2, 0, self.max_norm)
        return u

    def get_config(self):
        base_config = super().get_config()
        return {
            **base_config,
            "learning_rate": self._serialize_hyperparameter("learning_rate"),
        }



import tensorflow as tf

"""
Tensorflow Math functions with clipping as required for hyperoblic operations.
"""

def cosh(x, clamp=15):
    return tf.math.cosh(tf.clip_by_value(x, clip_value_min=-clamp, clip_value_max=clamp))

def sinh(x, clamp=15):
    return tf.math.sinh(tf.clip_by_value(x, clip_value_min=-clamp, clip_value_max=clamp))

def tanh(x, clamp=15):
    return tf.math.tanh(tf.clip_by_value(x, clip_value_min=-clamp, clip_value_max=clamp))

def arcosh(x):
    x = tf.clip_by_value(x, clip_value_min=1+1e-15, clip_value_max=tf.reduce_max(x))
    return tf.math.acosh(x)

def asinh(x):
    return tf.math.asinh(x)

def atanh(x):
    x = tf.clip_by_value(x, clip_value_min=-1 + 1e-15, clip_value_max=1 - 1e-15)
    return tf.math.atanh(x)

@tf.custom_gradient
def custom_artanh_cg(x):
    x = tf.clip_by_value(x, clip_value_min=-1 + 1e-15, clip_value_max=1 - 1e-15)
    z = tf.cast(x, tf.float64, name=None)
    temp = tf.math.subtract(tf.math.log(1 + z), (tf.math.log(1 - z)))
    def artanh_grad(grad):
        return grad/ (1 - x ** 2)

    return tf.cast(tf.math.multiply(temp, 0.5), x.dtype, name=None), artanh_grad



'''
This package contains linear algebra utils. 
Functions with suffix "_mp" operate on mpmath matrices. 
'''
import mpmath as mpm
import numpy as np
from math import sqrt


def rotate_3D(x, y):
    '''Returns 3D rotation matrix that sends x to the direction parallel to y'''
    xnorm = np.linalg.norm(x)
    ynorm = np.linalg.norm(y)
    cos = np.dot(x,y)/(xnorm*ynorm)
    sin = sqrt(1. - cos**2)
    K = (np.outer(y,x) - np.outer(x,y))/(xnorm*ynorm*sin)
    return np.eye(3) + sin*K + (1.-cos)*(K@K)


def cross_mp(x, y):
    return mpm.matrix([
        x[1]*y[2]-x[2]*y[1],
        x[2]*y[0]-x[0]*y[2],
        x[0]*y[1]-x[1]*y[0],
    ])


def rotate_3D_mp(x, y):
    xnorm = mpm.norm(x)
    ynorm = mpm.norm(y)
    cos = mpm.fdot(x,y)/(xnorm*ynorm)
    sin = mpm.sqrt(1.-cos**2)
    K = (y.T*x-x.T*y)/(xnorm*ynorm*sin)
    return mpm.eye(3) + sin*K + (1.-cos)*(K*K)


def rotate_mp(pts, x, y):
    out = mpm.zeros(pts.rows, pts.cols)
    v = x/mpm.norm(x) 
    cos = mpm.fdot(x,y) / (mpm.norm(x), mpm.norm(y))
    sin = mpm.sqrt(1.-cos**2)
    u = y - mpm.fdot(v, y) * v 

    mat = mpm.eye(x.cols) - u.T * u - v.T * v \
        + cos * u.T * u - sin * v.T * u + sin * u.T * v + cos * v.T * v
    return mat


def from_numpy(M):
    '''Convert 2D numpy array to mpmath matrix'''
    N = mpm.matrix(*M.shape)
    n, m = M.shape
    for i in range(n):
        for j in range(m):
            N[i,j] = M[i,j]
    return N


def to_numpy(N):
    '''Convert 2D mpmath matrix to numpy array'''
    M = np.zeros((N.rows,N.cols), dtype=np.float64)
    for i in range(N.rows):
        M[i,:] = N[i,:]
    return M



from math import sqrt, floor, log, log2
from itertools import product
import networkx as nx
from networkx.algorithms.shortest_paths.weighted import single_source_dijkstra_path_length
import numpy as np
import mpmath as mpm

from ..utils.multiprecision import poincare_reflect0
from ..utils.linalg import rotate_3D_mp

def sarkar_embedding(tree, root, **kwargs):
    ''' 
    Embed a tree in the Poincare disc using Sarkar's algorithm 
    from "Low Distortion Delaunay Embedding of Trees in Hyperbolic Plane.
        Args:
            tree (networkx.Graph) : The tree represented with int node labels.
                  Weighted trees should have the edge attribute "weight"
            root (int): The node to use as the root of the embedding 
        Keyword Args:
            weighted (bool): True if the tree is weighted (default True)
            tau (float): the scaling factor for distances. 
                        By default it is calculated based on statistics of the tree.
            epsilon (float): parameter >0 controlling distortion bound (default 0.1).
            precision (int): number of bits of precision to use.
                            By default it is calculated based on tau and epsilon.
        Returns:
            size N x 2 mpmath.matrix containing the coordinates of embedded nodes
    '''
    eps = kwargs.get("epsilon",0.1)
    weighted = kwargs.get("weighted", True)
    tau = kwargs.get("tau")
    max_deg = max(tree.degree)[1]

    if tau is None:
        tau = (1+eps)/eps * mpm.log(2*max_deg/ mpm.pi)
    prc = kwargs.get("precision")
    if prc is None:
        prc = _embedding_precision(tree,root,eps)
    mpm.mp.dps = prc
    
    n = tree.order()
    emb = mpm.zeros(n,2)
    place = []

    # place the children of root
    for i, v in enumerate(tree[root]):
        if weighted: 
            r = mpm.tanh( tau*tree[root][v]["weight"])
        else:
            r = mpm.tanh(tau)
        theta = 2*i*mpm.pi / tree.degree[root]
        emb[v,0] = r*mpm.cos(theta)
        emb[v,1] = r*mpm.sin(theta)
        place.append((root,v))
    
    # TODO parallelize this
    while place:
        u, v = place.pop() # u is the parent of v
        p, x = emb[u,:], emb[v,:]
        rp = poincare_reflect0(x, p, precision=prc)
        arg = mpm.acos(rp[0]/mpm.norm(rp))
        if rp[1] < 0:
            arg = 2*mpm.pi - arg
            
        theta = 2*mpm.pi / tree.degree[v]
        i=0
        for w in tree[v]:
            if w == u: continue
            i+=1
            if weighted:
                r = mpm.tanh(tau*tree[v][w]["weight"])
            else:
                r = mpm.tanh(tau)
            w_emb = r * mpm.matrix([mpm.cos(arg+theta*i),mpm.sin(arg+theta*i)]).T
            w_emb = poincare_reflect0(x, w_emb, precision=prc)
            emb[w,:] = w_emb
            place.append((v,w))
    return emb

def sarkar_embedding_3D(tree, root, **kwargs):
    eps = kwargs.get("eps",0.1)
    weighted = kwargs.get("weighted", True)
    tau = kwargs.get("tau")
    max_deg = max(tree.degree)[1]

    if tau is None:
        tau = (1+eps)/eps * mpm.log(2*max_deg/ mpm.pi)
    prc = kwargs.get("precision")
    if prc is None:
        prc = _embedding_precision(tree,root,eps)
    mpm.mp.dps = prc
    
    n = tree.order()
    emb = mpm.zeros(n,3)
    place = []

    # place the children of root
    fib = fib_2D_code(tree.degree[root])
    for i, v in enumerate(tree[root]):
        r = mpm.tanh(tau*tree[root][v].get("weight",1.))
        v_emb = r * mpm.matrix([[fib[i,0],fib[i,1],fib[i,2]]])
        emb[v,:]=v_emb
        place.append((root,v))
    
    while place:
        u, v = place.pop() # u is the parent of v
        u_emb, v_emb = emb[u,:], emb[v,:]

        # reflect and rotate so that embedding(v) is at (0,0,0) 
        # and embedding(u) is in the direction of (0,0,1)
        u_emb = poincare_reflect0(v_emb, u_emb, precision=prc)
        R = rotate_3D_mp(mpm.matrix([[0.,0.,1.]]), u_emb)
        #u_emb = (R.T * u_emb).T

        # place children of v 
        fib = fib_2D_code(tree.degree[v])
        i=0
        for w in tree[v]:
            if w == u: # i=0 is for u (parent of v)
                continue            
            i+=1
            r = mpm.tanh(tau*tree[w][v].get("weight",1.))
            w_emb = r * mpm.matrix([[fib[i,0],fib[i,1],fib[i,2]]])

            #undo reflection and rotation
            w_emb = (R * w_emb.T).T
            w_emb = poincare_reflect0(v_emb, w_emb, precision=prc)
            emb[w,:] = w_emb
            place.append((v,w))
    return emb

def _embedding_precision(tree, root, eps):
    dists = single_source_dijkstra_path_length(tree,root)
    max_deg = max(tree.degree)[1]
    l = 2*max(dists.values())
    prc = floor( (log(max_deg+1)) * l/eps +1)
    return prc

def hadamard_code(n):
    ''' 
    Generate a binary Hadamard code 
        Args:
            n (int): if n = 2**k + r < 2**(k+1), generates a [n, k, 2**(k-1)] Hadamard code
        Returns:
            2**k x n np.array where code(i) = the ith row
    '''
    k = floor( log2(n) )
    r = n - 2**k 
    gen = np.zeros((k,2**k), dtype=np.int32)
    for i, w in enumerate( product([0,1], repeat=k) ):
        gen[:,i] = w
    code = gen.T @ gen % 2
    if r > 0: 
        code = np.concatenate( (code, code[:,:3]), axis=1 )
    return code

def fib_2D_code(n, eps=None):
    epsilons = [0.33, 1.33, 3.33, 10, 27, 75]
    if eps is None:
        if n < 24:
            eps = epsilons[0]
        elif n < 177:
            eps = epsilons[1]
        elif n < 890:
            eps = epsilons[2]
        elif n < 11000:
            eps = epsilons[3]
        elif n < 400000:
            eps = epsilons[4]
        else:
            eps = epsilons[5]

    golden = (1+sqrt(5))/2
    i = np.arange(0,n)
    theta = 2*np.pi*i / golden
    phi = np.arccos(1 - 2*(i+eps)/(n-1+2*eps))
    M = np.stack([np.cos(theta)*np.sin(phi), np.sin(theta)*np.sin(phi), np.cos(phi)], axis=1)

    # rotate so that first point is (0,0,1)
    cos = M[0,2]
    sin = sqrt(1-cos**2)
    R = np.array([
        [cos, 0., sin],
        [0. , 1., 0. ],
        [-sin,0., cos],
    ])
    return M@R



from math import sqrt, floor
from scipy.spatial.distance import squareform
import numpy as np
import mpmath as mpm

from ..utils.graph import to_networkx
from .metric import is_metric
import __hyperlib_embedding

def treerep(dists, **kwargs):
    """
    TreeRep algorithm from Sonthalia & Gilbert, 'Tree! I am no Tree! I am a Low Dimensional Hyperbolic Embedding'
    takes a metric (distance matrix) and computes a weighted tree that approximates it.
    	Args:
    		metric (ndarray): size NxN distance matrix or 
                            compressed matrix of length N*(N-1)//2 ( e.g from scipy.pdist )
    		tol (double): tolerance for checking equalities (default=0.1)
            return_networkx (bool): return a networkx.Graph instead of edges (default=False)
    	Returns:
            A dict mapping edges (u,v), u<v to their edge weight
        
        Notes:
            - TreeRep is a randomized algorithm, so you can generate multiple trees and take the best.
            - TreeRep inserts new nodes (called Steiner nodes) labeled with ints >= N.
                They may or may not have an interpretation in terms of the data.
            - some edge weights may be 0. In that case you can either retract the edge
                or perturb it to a small positive number.
    """
    tol = kwargs.get("tol",0.1)
    if len(dists.shape) == 1:
        N = floor( (1+sqrt(1+8*len(dists)))/2 )
        assert N*(N-1)//2 == len(dists)
        W = __hyperlib_embedding.treerep(dists, N, tol)
    elif len(dists.shape) == 2:
        assert is_metric(dists)
        W = __hyperlib_embedding.treerep(squareform(dists), dists.shape[0], tol)
    else:
        raise ValueError("Invalid distance matrix")

    if kwargs.get("return_networkx", False):
        return to_networkx(W)
    return W



from setuptools import find_packages
from distutils.core import setup, Extension
import pybind11
from pybind11.setup_helpers import Pybind11Extension, build_ext
from glob import glob
import sys

__version__ = "0.0.4"
with open("README.md") as f:
    readme = f.read()

ext_modules = [ 
    Pybind11Extension("__hyperlib_embedding", sorted(glob("hyperlib/embedding/src/*.cc")),
        include_dirs = ["hyperlib/embedding/include", pybind11.get_include()],
        cxx_std=11, 
    )
]

setup(
    name="hyperlib",
    version=__version__,
    packages = find_packages(),
    setup_requires = ["pip>=21"],
    install_requires = [
        "numpy>=1.19.3", 
        "tensorflow>=2.0.0", 
        "scipy>=1.7.0", 
        "mpmath", 
        "networkx", 
        "pybind11>=2.7.0"] + (
        ["gmpy2"] if sys.platform != "win32" else []
    ),
    ext_modules=ext_modules,
    author="Nalex.ai",
    author_email="info@nalexai.com",
    description="Library that contains implementations of machine learning components in the hyperbolic space",
    long_description=readme,
    long_description_content_type="text/markdown",
    project_urls = {
        "Source" : "https://github.com/nalexai/hyperlib",
        "Issues" : "https://github.com/nalexai/hyperlib/issues"
    },
    license_files = "LICENSE",
    url = "https://github.com/nalexai/hyperlib",
    classifiers = [ 
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ], 
)



import tensorflow as tf
from ..utils.functional import tanh, atanh, asinh
from .base import Manifold

class Poincare(Manifold):

    """
    Implementation of the poincare manifold,. This class can be used for mathematical functions on the poincare manifold.
    """

    def __init__(self,):
        super(Poincare, self).__init__()
        self.name = "PoincareBall"
        self.min_norm = 1e-15
        self.eps = {tf.float32: 4e-3, tf.float64: 1e-5}

    def mobius_matvec(self, m, x, c):
        """
        Generalization for matrix-vector multiplication to hyperbolic space defined as
        math::
            M \otimes_c x = (1/\sqrt{c}) \tanh\left(
                \frac{\|Mx\|_2}{\|x\|_2}\tanh^{-1}(\sqrt{c}\|x\|_2)
            \right)\frac{Mx}{\|Mx\|_2}
        Args:
            m : Tensor for multiplication
            x : Tensor point on poincare ball
            c : Tensor of size 1 representing the hyperbolic curvature.
        Returns
            Mobius matvec result
        """

        sqrt_c = c ** 0.5
        x_norm = tf.norm(x, axis=-1, keepdims=True, ord=2)
        max_num = tf.math.reduce_max(x_norm)
        x_norm = tf.clip_by_value(
            x_norm, clip_value_min=self.min_norm, clip_value_max=max_num
        )
        mx = x @ m
        mx_norm = tf.norm(mx, axis=-1, keepdims=True, ord=2)
        max_num = tf.math.reduce_max(mx_norm)
        mx_norm = tf.clip_by_value(
            mx_norm, clip_value_min=self.min_norm, clip_value_max=max_num
        )

        res_c = (
            tanh(mx_norm / x_norm * atanh(sqrt_c * x_norm)) * mx / (mx_norm * sqrt_c)
        )
        cond = tf.reduce_prod(
            tf.cast((mx == 0), tf.uint8, name=None), axis=-1, keepdims=True
        )
        res_0 = tf.zeros(1, dtype=res_c.dtype)
        res = tf.where(tf.cast(cond, tf.bool), res_0, res_c)
        return res

    def expmap(self, u, p, c):
        sqrt_c = c ** 0.5
        u_norm = u.norm(dim=-1, p=2, keepdim=True).clamp_min(self.min_norm)
        second_term = (
            tanh(sqrt_c / 2 * self._lambda_x(p, c) * u_norm) * u / (sqrt_c * u_norm)
        )
        gamma_1 = self.mobius_add(p, second_term, c)
        return gamma_1

    def expmap0(self, u, c):
        """
        Hyperbolic exponential map at zero in the Poincare ball model.
          Args:
            u: tensor of size B x dimension representing tangent vectors.
            c: tensor of size 1 representing the hyperbolic curvature.
          Returns:
            Tensor of shape B x dimension.
          """
        sqrt_c = c ** 0.5
        max_num = tf.math.reduce_max(u)
        u_norm = tf.clip_by_value(
            tf.norm(u, axis=-1, ord=2, keepdims=True),
            clip_value_min=self.min_norm,
            clip_value_max=max_num,
        )
        gamma_1 = tf.math.tanh(sqrt_c * u_norm) * u / (sqrt_c * u_norm)
        return gamma_1

    def logmap0(self, p, c):
        """
        Hyperbolic logarithmic map at zero in the Poincare ball model.
        Args:
          p: tensor of size B x dimension representing hyperbolic points.
          c: tensor of size 1 representing the hyperbolic curvature.
        Returns:
          Tensor of shape B x dimension.
        """
        sqrt_c = c ** 0.5
        p_norm = tf.norm(p, axis=-1, ord=2, keepdims=True)
        max_num = tf.math.reduce_max(p_norm)
        p_norm = tf.clip_by_value(
            p_norm, clip_value_min=self.min_norm, clip_value_max=max_num
        )
        scale = 1.0 / sqrt_c * atanh(sqrt_c * p_norm) / p_norm
        return scale * p

    def proj(self, x, c):
        """
        Safe projection on the manifold for numerical stability. This was mentioned in [1]
        Args:
            x : Tensor point on the Poincare ball
            c : Tensor of size 1 representing the hyperbolic curvature.
        Returns:
            Projected vector on the manifold
        References:
            [1] Hyperbolic Neural Networks, NIPS2018
            https://arxiv.org/abs/1805.09112
        """

        x_for_norm = tf.norm(x, axis=-1, keepdims=True, ord=2)
        max_num = tf.math.reduce_max(x_for_norm)
        norm = tf.clip_by_value(
            x_for_norm, clip_value_min=self.min_norm, clip_value_max=max_num
        )
        maxnorm = (1 - self.eps[x.dtype]) / (c ** 0.5)  # tf.math.reduce_max(x)
        cond = norm > maxnorm
        projected = x / norm * maxnorm
        return tf.where(cond, projected, x)

    def mobius_add(self, x, y, c):
        """Element-wise Mobius addition.
      Args:
        x: Tensor of size B x dimension representing hyperbolic points.
        y: Tensor of size B x dimension representing hyperbolic points.
        c: Tensor of size 1 representing the absolute hyperbolic curvature.
      Returns:
        Tensor of shape B x dimension representing the element-wise Mobius addition
        of x and y.
      """
        cx2 = c * tf.reduce_sum(x * x, axis=-1, keepdims=True)
        cy2 = c * tf.reduce_sum(y * y, axis=-1, keepdims=True)
        cxy = c * tf.reduce_sum(x * y, axis=-1, keepdims=True)
        num = (1 + 2 * cxy + cy2) * x + (1 - cx2) * y
        denom = 1 + 2 * cxy + cx2 * cy2
        return self.proj(num / tf.maximum(denom, self.min_norm), c)


    def hyp_act(self, act, x, c_in, c_out):
        """Apply an activation function to a tensor in the hyperbolic space"""
        xt = act(self.logmap0(x, c=c_in))
        return self.proj(self.expmap0(xt, c=c_out), c=c_out)

    def single_query_attn_scores(self, key, query, c):
        """
        Arguments:
            key: Hyperbolic key with shape (batch, seq, hidden_dim)
            query: Hyperbolic query with shape (batch, hidden_dim)
        Returns:
            Scores as scalars in R with shape (batch,seq,1)
        """
        euclid_key = self.logmap0(key, c)
        euclid_query = self.logmap0(query, c)
        scores = tf.matmul(euclid_query.unsqueeze(-1), euclid_key, transpose_b=True)
        denom = tf.norm(euclid_key, keepdims=True, axis=-1)
        scores = (1. / denom) * scores
        return scores




import tensorflow as tf

class Manifold(object):
    """
    Abstract class to define operations on a manifold.
    """

    def __init__(self):
        super().__init__()
        self.name = "manifold"
        self.min_norm = 1e-15
        self.eps = {tf.float32: 4e-3, tf.float64: 1e-5}


    def expmap(self, u, p, c):
        raise NotImplementedError

    def expmap0(self, u, c):
        """
        Hyperbolic exponential map at zero.
          Args:
            u: tensor of size B x dimension representing tangent vectors.
            c: tensor of size 1 representing the hyperbolic curvature.
          Returns:
            Tensor of shape B x dimension.
          """
        raise NotImplementedError

    def logmap0(self, p, c):
        """
        Hyperbolic logarithmic map at zero.
        Args:
          p: tensor of size B x dimension representing hyperbolic points.
          c: tensor of size 1 representing the hyperbolic curvature.
        Returns:
          Tensor of shape B x dimension.
        """
        raise NotImplementedError

    def proj(self, x, c):
        """
        Safe projection on the manifold for numerical stability. This was mentioned in [1]

        Args:
            x : Tensor point on the Poincare ball
            c : Tensor of size 1 representing the hyperbolic curvature.

        Returns:
            Projected vector on the manifold

        References:
            [1] Hyperbolic Neural Networks, NIPS2018
            https://arxiv.org/abs/1805.09112
        """

        raise NotImplementedError

    def hyp_act(self, act, c_in, c_out):
        """Apply an activation function to a tensor in the hyperbolic space"""
        raise NotImplementedError
